//Generic class for extraction of function arguments
class bbArg
{
    public:
        bbArg(bool);
        virtual const char *parse(const char *)=0;
        virtual void *getData()=0;
    protected:
        bool lastArg;
};

bbArg::bbArg(bool last)
{
    lastArg=last;
}

typedef std::vector<bbArg *> bbArgList;

//Function for parsing arguments for all the functions
bool parse(const char *source,bbArgList &argList)
{
    const char *current;
    current=strchr(source,'(')+1;
    for (bbArgList::iterator i = argList.begin();i!=argList.end(); i++)
    {
        current=(*i)->parse(current);
        if (current==NULL)
            return false;
    }
    return true;
}

//String argument class
class bbString:public bbArg
{
    public:
        bbString(bool);
        virtual const char *parse(const char *);
        virtual void *getData();
    protected:
        char data[MAX_LINE_LENGTH];
};

bbString::bbString(bool last):bbArg(last)
{
}

const char *bbString::parse(const char *source)
{
    bool single=false;
    char *start=strchr(source,'"');
    if (!start) 
    {
        start=strchr(source,'\'');
        single=true;
    }
    if (start)
    {
        char *end = NULL;
        if (!single)
            end=strchr(start+1,'"');
        else
            end=strchr(start+1,'\'');
        if (end>start)
        {
            strncpy(data,start+1,end-start-1);
            data[end-start-1]=0;
            if (lastArg)
                return strchr(end,')');
            else
                return strchr(end,',')+1;
        }
    }
    return NULL;
}

void *bbString::getData()
{
    return data;
}

#define GET_STRING(x) (char *)(x->getData())

//Integer argument class
class bbInt:public bbArg
{
    public:
        bbInt(bool);
        virtual const char *parse(const char *);
        virtual void *getData();
    protected:
        int data;
};

bbInt::bbInt(bool last):bbArg(last)
{
}

const char *bbInt::parse(const char *source)
{
    data=atoi(source);
    if (lastArg)
        return strchr(source,')');
    else
        return strchr(source,',')+1;
}

void *bbInt::getData()
{
    return &data;
}

#define GET_INT(x) *((int *)x->getData())

//Long Integer argument class
class bbLong:public bbArg
{
    public:
        bbLong(bool);
        virtual const char *parse(const char *);
        virtual void *getData();
    protected:
        long data;
};

bbLong::bbLong(bool last):bbArg(last)
{
}

const char *bbLong::parse(const char *source)
{
    data=atol(source);
    if (lastArg)
        return strchr(source,')');
    else
        return strchr(source,',')+1;
}

void *bbLong::getData()
{
    return &data;
}

#define GET_LONG(x) *((long int *)x->getData())

//Boolean argument class
class bbBool:public bbArg
{
    public:
        bbBool(bool);
        virtual const char *parse(const char *);
        virtual void *getData();
    protected:
        bool data;
};

bbBool::bbBool(bool last):bbArg(last)
{
}

const char *bbBool::parse(const char *source)
{
    data=atoi(source);
    if (lastArg)
        return strchr(source,')');
    else
        return strchr(source,',')+1;
}

void *bbBool::getData()
{
    return &data;
}

#define GET_BOOL(x) *((bool *)x->getData())

//Handles sending of bro@ms
void bbSendBroam(const char *source, SocketPlugin &caller)
{
    SendMessage(GetBBWnd(), BB_BROADCAST, 0, (LPARAM)source);
}

//Catches repeat calls to authorize
void bbAuthCatch(const char *source, SocketPlugin &caller)
{
    char buf[MAX_LINE_LENGTH];
    if (caller.authorized)
        strcpy(buf,"AUTHORIZED");
    else
        strcpy(buf,"NOT AUTHORIZED");
    send(caller.connection,buf,strlen(buf)+1,0);
}

//WriteBool
void bbWriteBool(const char *source, SocketPlugin &caller)
{
    bbArgList argList;
    argList.push_back(new bbString(false));
    argList.push_back(new bbString(false));
    argList.push_back(new bbBool(true));
    if (!parse(source,argList))
        return;
    WriteBool(GET_STRING(argList[0]),GET_STRING(argList[1]),GET_BOOL(argList[2]));
}

//WriteInt
void bbWriteInt(const char *source, SocketPlugin &caller)
{
    bbArgList argList;
    argList.push_back(new bbString(false));
    argList.push_back(new bbString(false));
    argList.push_back(new bbInt(true));
    if (!parse(source,argList))
        return;
    WriteInt(GET_STRING(argList[0]),GET_STRING(argList[1]),GET_INT(argList[2]));
}

//WriteString
void bbWriteString(const char *source, SocketPlugin &caller)
{
    bbArgList argList;
    argList.push_back(new bbString(false));
    argList.push_back(new bbString(false));
    argList.push_back(new bbString(true));
    if (!parse(source,argList))
        return;
    WriteString(GET_STRING(argList[0]),GET_STRING(argList[1]),GET_STRING(argList[2]));
}

//WriteColor
void bbWriteColor(const char *source, SocketPlugin &caller)
{
    bbArgList argList;
    argList.push_back(new bbString(false));
    argList.push_back(new bbString(false));
    argList.push_back(new bbInt(true));
    if (!parse(source,argList))
        return;
    WriteColor(GET_STRING(argList[0]),GET_STRING(argList[1]),GET_INT(argList[2]));
}

//ReadBool
void bbReadBool(const char *source, SocketPlugin &caller)
{
    bbArgList argList;
    argList.push_back(new bbString(false));
    argList.push_back(new bbString(false));
    argList.push_back(new bbBool(true));
    if (!parse(source,argList))
        return;
    int data;
    data = ReadBool(GET_STRING(argList[0]),GET_STRING(argList[1]),GET_BOOL(argList[2]));
    char buf[MAX_LINE_LENGTH];
    itoa(data,buf,10);
    send(caller.connection,buf,strlen(buf)+1,0);
}

//ReadInt
void bbReadInt(const char *source, SocketPlugin &caller)
{
    bbArgList argList;
    argList.push_back(new bbString(false));
    argList.push_back(new bbString(false));
    argList.push_back(new bbInt(true));
    if (!parse(source,argList))
        return;
    int data;
    data = ReadInt(GET_STRING(argList[0]),GET_STRING(argList[1]),GET_INT(argList[2]));
    char buf[MAX_LINE_LENGTH];
    itoa(data,buf,10);
    send(caller.connection,buf,strlen(buf)+1,0);
}

//ReadString
void bbReadString(const char *source, SocketPlugin &caller)
{
    bbArgList argList;
    argList.push_back(new bbString(false));
    argList.push_back(new bbString(false));
    argList.push_back(new bbString(true));
    if (!parse(source,argList))
        return;
    const char *data;
    data = ReadString(GET_STRING(argList[0]),GET_STRING(argList[1]),GET_STRING(argList[2]));
    char buf[MAX_LINE_LENGTH];
    strcpy(buf,"\"");
    strcat(buf,data);
    strcat(buf,"\"");
    send(caller.connection,buf,strlen(buf)+1,0);
}

//ReadColor
void bbReadColor(const char *source, SocketPlugin &caller)
{
    bbArgList argList;
    argList.push_back(new bbString(false));
    argList.push_back(new bbString(false));
    argList.push_back(new bbString(true));
    if (!parse(source,argList))
        return;
    COLORREF data;
    data = ReadColor(GET_STRING(argList[0]),GET_STRING(argList[1]),GET_STRING(argList[2]));
    char buf[MAX_LINE_LENGTH];
    itoa(data,buf,10);
    send(caller.connection,buf,strlen(buf)+1,0);
}

//ReadValue
void bbReadValue(const char *source, SocketPlugin &caller)
{
    bbArgList argList;
    argList.push_back(new bbString(false));
    argList.push_back(new bbString(false));
    argList.push_back(new bbLong(true));
    if (!parse(source,argList))
        return;
    const char *data;
    if (GET_LONG(argList[2])==-1)
        data = ReadValue(GET_STRING(argList[0]),GET_STRING(argList[1]));
    else
        data = ReadValue(GET_STRING(argList[0]),GET_STRING(argList[1]),(long int *)argList[2]->getData());
    char buf[MAX_LINE_LENGTH];
    strcpy(buf,"<\"");
    if (data) strcat(buf,data);
    strcat(buf,"\" , ");
    char buf2[20];
    ltoa(GET_LONG(argList[2]),buf2,10);
    strcat(buf,buf2);
    strcat(buf,">");
    send(caller.connection,buf,strlen(buf)+1,0);
}

//GetDesktopInfo
void bbGetDesktopInfo(const char *source, SocketPlugin &caller)
{
    DesktopInfo deskinfo;
    GetDesktopInfo(&deskinfo);
    char buf[MAX_LINE_LENGTH];
    memset(buf,0,MAX_LINE_LENGTH);
    sprintf(buf,"<\"%s\",%i,%i,%i",deskinfo.name,deskinfo.isCurrent? 1 : 0,deskinfo.number,deskinfo.ScreensX);
    if (branch==BBLEAN)
    {
        string_node *sn=deskinfo.deskNames;
        for(int i=0;i<deskinfo.ScreensX;++i)
        {
            sprintf(buf+strlen(buf),",\"%s\"",sn->str);
            sn=sn->next;
        }
    }
    buf[strlen(buf)]='>';
    buf[strlen(buf)]=0;
    send(caller.connection,buf,strlen(buf)+1,0);
}

//GetBBVersion
void bbGetBBVersion(const char *source, SocketPlugin &caller)
{
    const char *data;
    data = GetBBVersion();
    char buf[MAX_LINE_LENGTH];
    strcpy(buf,"\"");
    strcat(buf,data);
    strcat(buf,"\"");
    send(caller.connection,buf,strlen(buf)+1,0);
}

//GetBBWnd()
void bbGetBBWnd(const char *source, SocketPlugin &caller)
{
    HWND data;
    data = GetBBWnd();
    char buf[100];
    ltoa((long int)data,buf,10);
    //std::stringstream buf;
    //buf<<data;
    //MessageBox(NULL,buf,buf,MB_OK);
    send(caller.connection,buf,strlen(buf)+1,0);
}

//GetOSInfo()
void bbGetOSInfo(const char *source, SocketPlugin &caller)
{
    const char *data;
    data = GetOSInfo();
    char buf[MAX_LINE_LENGTH];
    strcpy(buf,"\"");
    strcat(buf,data);
    strcat(buf,"\"");
    send(caller.connection,buf,strlen(buf)+1,0);
}

//stylePath
void bbStylePath(const char *source, SocketPlugin &caller)
{
    bbArgList argList;
    argList.push_back(new bbString(true));
    if (!parse(source,argList))
        return;
    const char *data;
    if (stricmp("NULL",GET_STRING(argList[0])))
        data=stylePath(GET_STRING(argList[0]));
    else
        data=stylePath(NULL);
    char buf[MAX_LINE_LENGTH];
    strcpy(buf,"\"");
    strcat(buf,data);
    strcat(buf,"\"");
    send(caller.connection,buf,strlen(buf)+1,0);
}

//---------------------------------------------------------
//Menu Functions
//---------------------------------------------------------
//MakeMenu
void bbMakeMenu(const char *source, SocketPlugin &caller)
{
    bbArgList argList;
    argList.push_back(new bbString(true));
    if (!parse(source,argList))
        return;
    int newID;
    if (caller.menus.empty())
        newID=0;
    else
        newID=(*(caller.menus.rbegin())).first+1;
    Menu *p=MakeMenu(GET_STRING(argList[0]));
    if (p)
    {
        caller.menus[newID]=p;
        char buf[MAX_LINE_LENGTH];
        sprintf(buf,"%i",newID);
        send(caller.connection,buf,strlen(buf)+1,0);
        return;
    }
    else
    {
        send(caller.connection, MNUERR_CREATE ,strlen( MNUERR_CREATE )+1,0);
        return;
    }
}

//MakeNamedMenu
void bbMakeNamedMenu(const char *source, SocketPlugin &caller)
{
    bbArgList argList;
    argList.push_back(new bbString(false));
    argList.push_back(new bbString(false));
    argList.push_back(new bbBool(true));
    if (!parse(source,argList))
        return;
    int newID;
    if (caller.menus.empty())
        newID=0;
    else
        newID=(*(caller.menus.rbegin())).first+1;
    Menu *p=MakeNamedMenu(GET_STRING(argList[0]),GET_STRING(argList[1]),GET_BOOL(argList[2]));
    if (p)
    {
        caller.menus[newID]=p;
        char buf[MAX_LINE_LENGTH];
        sprintf(buf,"%i",newID);
        send(caller.connection,buf,strlen(buf)+1,0);
        return;
    }
    else
    {
        send(caller.connection,MNUERR_CREATE,strlen(MNUERR_CREATE)+1,0);
        return;
    }
}

//MakeMenuNOP
void bbMakeMenuNOP(const char *source, SocketPlugin &caller)
{
    bbArgList argList;
    argList.push_back(new bbInt(false));
    argList.push_back(new bbString(true));
    if (!parse(source,argList))
        return;
    MITERATOR tempIt;
    tempIt=caller.menus.find(GET_INT(argList[0]));
    if (tempIt!=caller.menus.end())
    {
        MenuItem *p=MakeMenuNOP((*tempIt).second,GET_STRING(argList[1]));
        if (p)
        {
            return;
        }
        else
        {
            send(caller.connection,MNUERR_CREATEITEM,strlen(MNUERR_CREATEITEM)+1,0);
            return;
        }
    }
    else
    {
        send(caller.connection,MNUERR_NOMENU,strlen(MNUERR_NOMENU)+1,0);
        return;
    }
}

//MakeSubMenu
void bbMakeSubMenu(const char *source, SocketPlugin &caller)
{
    bbArgList argList;
    argList.push_back(new bbInt(false));
    argList.push_back(new bbInt(false));
    argList.push_back(new bbString(true));
    if (!parse(source,argList))
        return;
    MITERATOR tempIt;
    tempIt=caller.menus.find(GET_INT(argList[0]));
    if (tempIt!=caller.menus.end())
    {
        MITERATOR tempIt2;
        tempIt2=caller.menus.find(GET_INT(argList[1]));
        if (tempIt2!=caller.menus.end())
        {  
            MenuItem *p=MakeSubmenu((*tempIt).second,(*tempIt2).second,GET_STRING(argList[2]));
            if (p)
            {
                //caller.menus.erase(tempIt2);
                return;
            }
            else
            {
                send(caller.connection,MNUERR_CREATEITEM,strlen(MNUERR_CREATEITEM)+1,0);
                return;
            }
        }
        else
        {
            send(caller.connection,MNUERR_NOSUBMENU,strlen(MNUERR_NOSUBMENU)+1,0);
            return;
        }
    }
    else
    {
        send(caller.connection,MNUERR_NOMENU,strlen(MNUERR_NOMENU)+1,0);
        return;
    }
}

//ShowMenu
void bbShowMenu(const char *source, SocketPlugin &caller)
{
    bbArgList argList;
    argList.push_back(new bbInt(true));
    if (!parse(source,argList))
        return;
    MITERATOR tempIt;
    tempIt=caller.menus.find(GET_INT(argList[0]));
    if (tempIt!=caller.menus.end())
    {
        ShowMenu((*tempIt).second);
        //caller.menus.erase(tempIt);
        return;
    }
    else
    {
        send(caller.connection,MNUERR_NOMENU,strlen(MNUERR_NOMENU)+1,0);
        return;
    }
}

//MakeMenuItem
void bbMakeMenuItem(const char *source, SocketPlugin &caller)
{
    bbArgList argList;
    argList.push_back(new bbInt(false));
    argList.push_back(new bbString(false));
    argList.push_back(new bbString(false));
    argList.push_back(new bbBool(true));
    if (!parse(source,argList))
        return;
    MITERATOR tempIt;
    tempIt=caller.menus.find(GET_INT(argList[0]));
    if (tempIt!=caller.menus.end())
    {
        MenuItem *p=MakeMenuItem((*tempIt).second,GET_STRING(argList[1]),GET_STRING(argList[2]),GET_BOOL(argList[3]));
        if (p)
        {
            return;
        }
        else
        {
            send(caller.connection,MNUERR_CREATEITEM,strlen(MNUERR_CREATEITEM)+1,0);
            return;
        }
    }
    else
    {
        send(caller.connection,MNUERR_NOMENU,strlen(MNUERR_NOMENU)+1,0);
        return;
    }
}

//MakeMenuItemInt
void bbMakeMenuItemInt(const char *source, SocketPlugin &caller)
{
    bbArgList argList;
    argList.push_back(new bbInt(false));
    argList.push_back(new bbString(false));
    argList.push_back(new bbString(false));
    argList.push_back(new bbInt(false));
    argList.push_back(new bbInt(false));
    argList.push_back(new bbInt(true));
    if (!parse(source,argList))
        return;
    MITERATOR tempIt;
    tempIt=caller.menus.find(GET_INT(argList[0]));
    if (tempIt!=caller.menus.end())
    {
        MenuItem *p=MakeMenuItemInt((*tempIt).second,GET_STRING(argList[1]),GET_STRING(argList[2]),GET_INT(argList[3]),GET_INT(argList[4]),GET_INT(argList[5]));
        if (p)
        {
            return;
        }
        else
        {
            send(caller.connection,MNUERR_CREATEITEM,strlen(MNUERR_CREATEITEM)+1,0);
            return;
        }
    }
    else
    {
        send(caller.connection,MNUERR_NOMENU,strlen(MNUERR_NOMENU)+1,0);
        return;
    }
}
//MakeMenuItemString
void bbMakeMenuItemString(const char *source, SocketPlugin &caller)
{
    bbArgList argList;
    argList.push_back(new bbInt(false));
    argList.push_back(new bbString(false));
    argList.push_back(new bbString(false));
    argList.push_back(new bbString(true));
    if (!parse(source,argList))
        return;
    MITERATOR tempIt;
    tempIt=caller.menus.find(GET_INT(argList[0]));
    if (tempIt!=caller.menus.end())
    {
        MenuItem *p=MakeMenuItemString((*tempIt).second,GET_STRING(argList[1]),GET_STRING(argList[2]),GET_STRING(argList[3]));
        if (p)
        {
            return;
        }
        else
        {
            send(caller.connection,MNUERR_CREATEITEM,strlen(MNUERR_CREATEITEM)+1,0);
            return;
        }
    }
    else
    {
        send(caller.connection,MNUERR_NOMENU,strlen(MNUERR_NOMENU)+1,0);
        return;
    }
}

//DelMenu
void bbDelMenu(const char *source, SocketPlugin &caller)
{
    bbArgList argList;
    argList.push_back(new bbInt(true));
    if (!parse(source,argList))
        return;
    MITERATOR tempIt;
    tempIt=caller.menus.find(GET_INT(argList[0]));
    if (tempIt!=caller.menus.end())
    {
        DelMenu((*tempIt).second);
        caller.menus.erase(tempIt);
    }
    else
    {
        send(caller.connection,MNUERR_NOMENU,strlen(MNUERR_NOMENU)+1,0);
    }
}

//-----------------------------------------------------------
//BBSocket specific
//-----------------------------------------------------------
//Subscribe
void subscribe(std::vector<unsigned int> &current,unsigned int newSub, bool add)
{
    std::vector<unsigned int>::iterator check;
    if ((check=std::find(current.begin(),current.end(),newSub))==current.end())
    {
        if (add) current.push_back(newSub);
    }
    else
        if (!add) current.erase(check);
}

void bbSubscribe(const char *source, SocketPlugin &caller)
{
    bbArgList argList;
    argList.push_back(new bbString(true));
    if (!parse(source,argList))
        return;
    char *input=GET_STRING(argList[0]);
    if (!stricmp(input,"BB_BROADCAST"))
        subscribe(caller.subscriptions,BB_BROADCAST,true);
    else if (!stricmp(input,"BB_RECONFIGURE"))
        subscribe(caller.subscriptions,BB_RECONFIGURE,true);
    else if (!stricmp(input,"BB_DESKTOPINFO"))
        subscribe(caller.subscriptions,BB_DESKTOPINFO,true);
}

//Unsubscribe
void bbUnsubscribe(const char *source, SocketPlugin &caller)
{
    bbArgList argList;
    argList.push_back(new bbString(true));
    if (!parse(source,argList))
        return;
    char *input=GET_STRING(argList[0]);
    if (!stricmp(input,"BB_BROADCAST"))
        subscribe(caller.subscriptions,BB_BROADCAST,false);
    else if (!stricmp(input,"BB_RECONFIGURE"))
        subscribe(caller.subscriptions,BB_RECONFIGURE,false);
    else if (!stricmp(input,"BB_DESKTOPINFO"))
        subscribe(caller.subscriptions,BB_DESKTOPINFO,false);
}

/*
//This results in inabiity to load plugin
//GetBlackboxPath()
class bbGetBlackboxPath:public bbFunction
{
    public:
        bbGetBlackboxPath();
    protected:
        bool call(SOCKET);
};

bbGetBlackboxPath::bbGetBlackboxPath():bbFunction("GetBlackboxPath")
{
}

bool bbGetBlackboxPath::call(SOCKET caller)
{
    char data[MAX_PATH];
    GetBlackboxPath(data,MAX_PATH-1);
    send(caller,data,strlen(data)+1,0);
    return true;
}
*/
void initFunctionList()
{
    functionMap["GetBBVersion("]=bbGetBBVersion;
    functionMap["GetBBWnd("]=bbGetBBWnd;
    functionMap["GetOSInfo("]=bbGetOSInfo;
    functionMap["GetDesktopInfo("]=bbGetDesktopInfo;
    functionMap["MakeMenu("]=bbMakeMenu;
    functionMap["MakeNamedMenu("]=bbMakeNamedMenu;
    functionMap["ShowMenu("]=bbShowMenu;
    functionMap["MakeMenuNOP("]=bbMakeMenuNOP;
    functionMap["MakeSubMenu("]=bbMakeSubMenu;
    functionMap["MakeMenuItem("]=bbMakeMenuItem;
    functionMap["MakeMenuItemInt("]=bbMakeMenuItemInt;
    functionMap["MakeMenuItemString("]=bbDelMenu;
    functionMap["DelMenu("]=bbDelMenu;
    
    functionMap["stylePath("]=bbStylePath;
    
    functionMap["ReadBool("]=bbReadBool;
    functionMap["ReadInt("]=bbReadInt;
    functionMap["ReadString("]=bbReadString;
    functionMap["ReadColor("]=bbReadColor;
    functionMap["ReadValue("]=bbReadValue;
    functionMap["WriteBool("]=bbWriteBool;
    functionMap["WriteInt("]=bbWriteInt;
    functionMap["WriteString("]=bbWriteString;
    functionMap["WriteColor("]=bbWriteColor;
    //functionList.push_back(new bbGetBlackboxPath);
    
    //BBSocket Specific
    functionMap["Subscribe("]=bbSubscribe;
    functionMap["Unsubscribe("]=bbUnsubscribe;
    functionMap["AUTHORIZE "]=bbAuthCatch;
    functionMap["@"]=bbSendBroam;
    
}
