#define WM_WSAASYNC (WM_USER +5)

#include <bbLean/BBApi.h>
#include <winsock.h>
#include <vector>
#include <map>
#include <algorithm>

enum {BBLEAN,BB4WIN,XOBLITE};

// Global vars
int branch=0;
char rcpath[MAX_PATH];
bool localHostOnly;
char bindTo[100];
bool requireAuth;
bool listenOnStart;
char *findHelp;
int listenPort;
HINSTANCE hInstance;
static HANDLE   hProcess = NULL;
static HWND     hPluginWnd = NULL;
static int      bb_messages[] = {BB_BROADCAST, BB_RECONFIGURE, BB_DESKTOPINFO, 0};
//Program messages
const char MNUERR_CREATE[] = "ERROR: Menu creation failed";
const char MNUERR_NOMENU[] = "ERROR: Invalid Menu ID";
const char MNUERR_NOSUBMENU[] = "ERROR: Invalid Submenu ID";
const char MNUERR_CREATEITEM[] = "ERROR: Menu Item creation failed";
const char AUTH_MESSAGE[] = "A process is attempting to communicate with bbSocket:\r\n%s\r\n\r\nYou will only be prompted once, so please select carefully!\r\nDo you wish to allow this?";

void ReadRCSettings(void);

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
//Plugin Classes
//---------------------------------------------------------------------------------
//ListenSocket : Encapsulates sockets listening for connections
//Definition
class ListenSocket
{
    public:
        ListenSocket(const char *, bool start=true);
        ~ListenSocket();
        void Listen(bool);
        void ListenToggle();
    private:
        SOCKET connection;
        char name[100];
        int port;
        char bindTo[100];
};

//Implementation
ListenSocket::ListenSocket(const char *source, bool start)
{
    connection=(SOCKET)NULL;
    strcpy(name,source);
    const char *delim=strchr(source,':');
    bindTo[0]=0;
    if (delim) //A port was specified
    {
        if (delim!=source) //Binding address specified
        {
            strncpy(bindTo,source,delim-source);
            bindTo[delim-source]=0;
        }
        port=atoi(delim+1);
    }
    else
    {
        strcpy(bindTo,source); //Entire string is address
        port=listenPort; //Use defualt port
    }
    if (start) Listen(true);
}

ListenSocket::~ListenSocket()
{
    closesocket(connection);
}

void ListenSocket::Listen(bool on)
{
    if ((on)&&(!connection))
    {
        connection = socket(AF_INET, SOCK_STREAM, 0);
        sockaddr_in sa;
        memset(&sa, 0, sizeof(sa));
        sa.sin_family = PF_INET;             
        sa.sin_port = htons(port);
        if (strlen(bindTo))
        {
            hostent* host_ent = gethostbyname(bindTo);
		  memcpy((char *)&sa.sin_addr, host_ent->h_addr, host_ent->h_length);
        }
	
        if(bind(connection, (sockaddr *)&sa, sizeof(sockaddr_in))==SOCKET_ERROR)
            MessageBox(hPluginWnd,"Could not bind to specified Address and Port","BBSocket Error",MB_OK);
                  
        BOOL bOptVal = TRUE;
	   int bOptLen = sizeof(BOOL);
	   setsockopt(connection, SOL_SOCKET, SO_KEEPALIVE,(char*)&bOptVal, bOptLen);
	
        WSAAsyncSelect(connection, hPluginWnd, WM_WSAASYNC, FD_READ | FD_WRITE |
                                              FD_ACCEPT | FD_CLOSE);
        listen(connection,10);
    }
    else if (connection)
    {
        closesocket(connection);
        connection=(SOCKET)NULL;
    }
}

void ListenSocket::ListenToggle()
{
    if (connection) 
        Listen(false);
    else
        Listen(true);
}

//---------------------------------------------------------------------------------
//SocketPlugin : Encapsulates the connection to a socket plugin
//Definition
class SocketPlugin
{
    public:
        SocketPlugin(SOCKET);
        ~SocketPlugin();
        void authorize(char *);
        const char*getName() {return name;}
        std::map<int,Menu *> menus;
        std::vector<unsigned int> subscriptions;
        SOCKET connection;
        int authorized;
        
    protected:
        char name[MAX_LINE_LENGTH];
    private:         
};


//Implementation
SocketPlugin::SocketPlugin(SOCKET con)
{
    connection=con;
    strcpy(name,"Unidentified");
    if (requireAuth)
        authorized = -1;
    else
        authorized = 1;
    subscriptions.push_back(BB_BROADCAST);
}

SocketPlugin::~SocketPlugin()
{
    while(!menus.empty())
    {
        DelMenu((*menus.begin()).second);
        menus.erase(menus.begin());
    }
    closesocket(connection);
}

void SocketPlugin::authorize(char *plugName)
{
    if (!strcmp("Unidentified",plugName))
        return;
    strcpy(name,plugName);
    char rc[MAX_LINE_LENGTH];
	char *msg = new char[strlen(name) + strlen(AUTH_MESSAGE) + 1];
	strcpy(rc, "bbSocket.auth.");
	strncat(rc, name, MAX_LINE_LENGTH - strlen(rc) - 2);
	strcat(rc, ":");
	authorized = ReadInt(rcpath, rc, -1);
	if (authorized == -1)
	{
		sprintf(msg, AUTH_MESSAGE, name);
		if (MessageBox(hPluginWnd, msg, "bbSocket Plugin Authorization", MB_YESNO | MB_ICONQUESTION | MB_SETFOREGROUND | MB_TOPMOST) == IDYES)
			authorized = TRUE;
		else
			authorized = FALSE;
        WriteInt(rcpath, rc, authorized);
	}
}


//---------------------------------------------------------------------------------
//Maps used by the plugin
//Comparison object used by bro@m and function maps
class strCmp 
{
    public:
        bool operator()( const char* s1, const char* s2 ) const 
        {
            return strnicmp( s1, s2 ,s2==findHelp?strlen(s1):strlen(s2)) < 0;
        }
};

//Maps bro@m and function names to functions to be executed
std::map<const char *,void (*)(const char *,SocketPlugin&),strCmp> functionMap;
std::map<const char *,void (*)(const char *),strCmp> broamMap;

//---------------------------------------------------------------------------------
//Variable, types and functions for the plugin classes

//Iterator type for use with SocketPlugin menu implementation
typedef std::map<int , Menu *>::iterator MITERATOR;

//Include the defnitions for all functions and arguments
#include "bbAPIfuncs.h"

//Vecctors to hold the socket classes
std::vector<ListenSocket *> ListenList;
std::vector<SocketPlugin *> SocketPluginList;

//Iterator types for the vectors
typedef std::vector<ListenSocket *>::iterator LSITERATOR;
typedef std::vector<SocketPlugin *>::iterator SPITERATOR;

//Used for locating a plugin instance by socket handle
SPITERATOR FindIteratorBySocket(SOCKET s)
{
    for (SPITERATOR i = SocketPluginList.begin(); i != SocketPluginList.end(); i++)
		if ((*i)->connection == s)
			return i;
	return SocketPluginList.end();
}


//---------------------------------------------------------------------------------
//Functions which implement the plugin bro@ms
//@BBSocket.Listen
void broamListen(const char *broam)
{
    if (strlen(broam)>=18)
    {
        if ((broam[17]=='t')||(broam[17]=='T'))
            for(LSITERATOR i = ListenList.begin();i!=ListenList.end();++i)
                (*i)->Listen(true);
        else if ((broam[17]=='f')||(broam[17]=='F'))
            for(LSITERATOR i = ListenList.begin();i!=ListenList.end();++i)
                (*i)->Listen(false);
    }
    else
        for(LSITERATOR i = ListenList.begin();i!=ListenList.end();++i)
            (*i)->ListenToggle();
}

//@BBSocket.KillAll
void broamKillAll(const char *broam)
{
    while(!SocketPluginList.empty())
    {
        SPITERATOR plug = SocketPluginList.begin();
        closesocket((*plug)->connection);
        SocketPluginList.erase(plug);
    }
}

//@BBSocket.Kill
void broamKill(const char *broam)
{
    const char *name=broam+15;
    for (SPITERATOR i = SocketPluginList.begin(); i != SocketPluginList.end(); i++)
		if (!stricmp((*i)->getName(),name))
		{
            closesocket((*i)->connection);
            SocketPluginList.erase(i);
            break;   
        }       
}

//Bro@m Map initialization function
void InitBroamMap()
{
    broamMap["@BBSocket.Listen"]=broamListen;
    broamMap["@BBSocket.KillAll"]=broamKillAll;
    broamMap["@BBSocket.Kill "]=broamKill;
}

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
//Window procedure
LRESULT CALLBACK PluginProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    char buf[MAX_LINE_LENGTH];
    int num;
    std::vector<unsigned int>::iterator check;
    switch(message)
    {
        case BB_BROADCAST:
        {
            findHelp=(char *)lParam;
            std::map<const char *,void (*)(const char *),strCmp>::iterator found;
            if ((found=broamMap.find((char *)lParam))!=broamMap.end())
                (*found).second((char *)lParam);
            for (SPITERATOR i = SocketPluginList.begin();i!=SocketPluginList.end(); i++)
                if (((*i)->authorized==TRUE)&&((check=std::find((*i)->subscriptions.begin(),(*i)->subscriptions.end(),message))!=(*i)->subscriptions.end()))
                    send((*i)->connection,(char*)lParam,strlen((char*)lParam)+1,0);
            return 0;
            break;
        }
        case BB_RECONFIGURE:
            for (SPITERATOR i = SocketPluginList.begin();i!=SocketPluginList.end(); i++)
                if (((*i)->authorized==TRUE)&&((check=std::find((*i)->subscriptions.begin(),(*i)->subscriptions.end(),message))!=(*i)->subscriptions.end()))
                    send((*i)->connection,"[BB_RECONFIGURE]",strlen("[BB_RECONFIGURE]")+1,0);
            for(LSITERATOR i = ListenList.begin();i!=ListenList.end();++i)
                delete (*i);
            ListenList.clear();
            ReadRCSettings();
            //StartListen();
            return 0;
            break;
        case BB_DESKTOPINFO:
            {
            DesktopInfo deskinfo;
            GetDesktopInfo(&deskinfo);
            memset(buf,0,MAX_LINE_LENGTH);
            sprintf(buf,"[BB_DESKTOPINFO <\"%s\",%i,%i,%i",deskinfo.name,deskinfo.isCurrent? 1 : 0,deskinfo.number,deskinfo.ScreensX);
            if (branch=BBLEAN)
            {
                string_node *sn=deskinfo.deskNames;
                for(int i=0;i<deskinfo.ScreensX;++i)
                {
                    sprintf(buf+strlen(buf),",\"%s\"",sn->str);
                    sn=sn->next;
                }
            }
            buf[strlen(buf)]='>';
            buf[strlen(buf)]=']';
            buf[strlen(buf)]=0;
            for (SPITERATOR i = SocketPluginList.begin();i!=SocketPluginList.end(); i++)
                if (((*i)->authorized==TRUE)&&((check=std::find((*i)->subscriptions.begin(),(*i)->subscriptions.end(),message))!=(*i)->subscriptions.end()))
                    send((*i)->connection,buf,strlen(buf)+1,0);
            }
            return 0;
            break;
        case WM_WSAASYNC:
             switch(WSAGETSELECTEVENT(lParam))
             {
                     case FD_ACCEPT:
                        SocketPluginList.push_back(new SocketPlugin(accept(wParam,0,0)));
						break;
                    case FD_READ:
                    {
                        SPITERATOR currentIterator = FindIteratorBySocket((SOCKET)wParam);
                        if (currentIterator == SocketPluginList.end())
                        {
                            num = recv(wParam, buf, MAX_LINE_LENGTH - 1, 0);
							buf[num] = 0;
                            //this should never happen!!!
							break;
						}

						SocketPlugin *currentPointer = *currentIterator;
						num = recv(currentPointer->connection, buf, MAX_LINE_LENGTH - 1, 0);
						buf[num] = 0;

						if (currentPointer->authorized == -1)
						{
							if (!strnicmp(buf, "AUTHORIZE ", strlen("AUTHORIZE ")))
							{
								char *p = buf + strlen("AUTHORIZE ");
								currentPointer->authorize(p);
							}
							else
                                break;
						}
						else if (currentPointer->authorized == FALSE)
						{
							send(currentPointer->connection, "NOT AUTHORIZED", strlen("NOT AUTHORIZED") + 1, 0);
							break;
						}
                        //if (buf[0]=='@') SendMessage(GetBBWnd(), BB_BROADCAST, 0, (LPARAM)buf);
                        findHelp=buf;
                        std::map<const char *,void (*)(const char *,SocketPlugin&),strCmp>::iterator found;
                        if ((found=functionMap.find(buf))!=functionMap.end())
                            (*found).second(buf,*currentPointer);
                        break;
                    }
                     case FD_WRITE:
                          break;
                     case FD_CLOSE:
                            SPITERATOR currentIterator = FindIteratorBySocket((SOCKET)wParam);
                            if (currentIterator!=SocketPluginList.end())
                                SocketPluginList.erase(currentIterator);
                            closesocket(wParam);
                            break;
             }
             return 0;
             break;
        
    }
    return DefWindowProc(hwnd, message, wParam, lParam);
}

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
//Blackbox-plugin interface

int beginPluginEx(HINSTANCE h_instance, HWND hSlit)
{
    return beginPlugin(h_instance);
}

//---------------------------------------------------------------------------------

int beginSlitPlugin(HINSTANCE h_instance, HWND hSlit)
{
    return beginPlugin(h_instance);
}

//---------------------------------------------------------------------------------

int beginPlugin(HINSTANCE h_instance)
{
    hInstance=h_instance;
    if (!strnicmp(GetBBVersion(),"bblean",6))
        branch=BBLEAN;
    else if (!strnicmp(GetBBVersion(),"0.",2))
        branch=BB4WIN;
    else
        branch=XOBLITE;
    WSADATA init;
    WSAStartup(MAKEWORD(1,1),&init);
    ReadRCSettings();
    WNDCLASS wc;
    ZeroMemory(&wc, sizeof(wc));
    wc.lpfnWndProc = PluginProc;
    wc.hInstance = h_instance;
    wc.lpszClassName = pluginInfo(PLUGIN_NAME);
    RegisterClass(&wc);

    hPluginWnd = CreateWindowEx(
        WS_EX_TOOLWINDOW,
        pluginInfo(PLUGIN_NAME),
        NULL,
        WS_POPUP|WS_ICONIC,
        -4, -4, 0, 0,
        HWND_MESSAGE,
        NULL,
        h_instance,
        NULL
    );

    hProcess = GetCurrentProcess();
    SendMessage(GetBBWnd(), BB_REGISTERMESSAGE, (WPARAM)hPluginWnd, (LPARAM)bb_messages);
    //if (listenOnStart) StartListen();
    initFunctionList();
    InitBroamMap();
    return 0;
}

//---------------------------------------------------------------------------------

void endPlugin(HINSTANCE h_instance)
{
    DestroyWindow(hPluginWnd);
    for(LSITERATOR i = ListenList.begin();i!=ListenList.end();++i)
        delete (*i);
    for (SPITERATOR i =SocketPluginList.begin();i!=SocketPluginList.end(); i++)
        delete (*i);
    SendMessage(GetBBWnd(), BB_UNREGISTERMESSAGE, (WPARAM)hPluginWnd, (LPARAM)bb_messages);
    UnregisterClass(pluginInfo(PLUGIN_NAME), h_instance);
}

//---------------------------------------------------------------------------------
//Define argument names for PluginInfo (These match the defnintions from xoblites bbapi.h)
#define PLUGIN_NAME				1
#define PLUGIN_VERSION			2
#define PLUGIN_AUTHOR			3
#define PLUGIN_RELEASEDATE		4
#define PLUGIN_LINK				5
#define PLUGIN_EMAIL			6
#define PLUGIN_BROAMS			7
#define PLUGIN_UPDATE_URL		8

LPCSTR    pluginInfo(int x)
{
    switch (x)
    {
        case PLUGIN_NAME:           return "bbSocket";
        case PLUGIN_VERSION:        return "0.0.5";
        case PLUGIN_AUTHOR:         return "Carsomyr";
        case PLUGIN_RELEASEDATE:    return "2006/01/14";
        case PLUGIN_LINK:           return "http://www.lostinthebox.com/viewtopic.php?t=2254";
        case PLUGIN_EMAIL:          return "s23120712@tuks.co.za";
        case PLUGIN_BROAMS:         return  "@BBSocket.KillAll" 
                                            "@BBSocket.Listen False"
                                            "@BBSocket.Listen True"
                                            "@BBSocket.Listen";
        default:                    return "bbSocket 0.0.5";
    }
}

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
//Function to read the RC fille
void ReadRCSettings(void)
{
    char name[MAX_PATH];
	int i = 0;
	do
	{
		// First and third, we look for the config file
		// in the same folder as the plugin...
		HINSTANCE hInst = hInstance;
		//Allow us to get the name of dll used
		GetModuleFileName(hInst, rcpath, sizeof(rcpath));
		char *plugin_name = strrchr(rcpath, '\\');
		plugin_name++;
		char *extension_start = strrchr(plugin_name, '.');
		extension_start++;
		extension_start[0]='r';
        extension_start[1]='c';
        extension_start[2]=0;
        strcpy(name,plugin_name);
		// second we check the blackbox directory
		if (1 == i) hInst = NULL;

		GetModuleFileName(hInst, rcpath, sizeof(rcpath));
		char *file_name_start = strrchr(rcpath, '\\');
		
		if (file_name_start) ++file_name_start;
		else file_name_start = strchr(rcpath, 0);
		strcpy(file_name_start, name);
	} while (++i < 3 && false == FileExists(rcpath));
    
	// If a config file was found we read the plugin settings from it...
	// ...if not, the ReadXXX functions give us just the defaults.
    
    listenPort  = ReadInt(rcpath, "bbSocket.port:", 25225);
    localHostOnly = ReadBool(rcpath, "bbSocket.localHostOnly:", true);
    requireAuth = ReadBool(rcpath, "bbSocket.requireAuth:", true);
    listenOnStart = ReadBool(rcpath, "bbsocket.listenOnStart:", true);
    strcpy(bindTo,ReadString(rcpath, "bbSocket.bind:", ""));
    const char *data;
    long index=0;
    do
    {
        data=ReadValue(rcpath, "bbSocket.bind:", &index);
        if (data)
        {
            if (strlen(data)) 
                ListenList.push_back(new ListenSocket(data,listenOnStart));
        }
        if (branch==XOBLITE) ++index;
    }
    while(data);
    if (ListenList.size()==0)
    {
        char bindstring[100];
        if (strlen(bindTo))
            sprintf(bindstring,"%s:%i",bindTo,listenPort);
        else
            if (localHostOnly)
                sprintf(bindstring,"127.0.0.1:%i",listenPort);
            else
                sprintf(bindstring,":%i",listenPort);
        ListenList.push_back(new ListenSocket(bindstring,listenOnStart));
    }
}
